@@ -143,6 +143,17 @@ public:
    */
   static PRUint32 GetAccessKeyFor(nsIContent *aContent);
 
+  /**
+   * Return DOM element related with the given node, i.e.
+   * a) itself if it is DOM element
+   * b) parent element if it is text node
+   * c) body element if it is HTML document node
+   * d) document element if it is document node.
+   *
+   * @param aNode  [in] the given DOM node
+   */
+  static already_AddRefed<nsIDOMElement> GetDOMElementFor(nsIDOMNode *aNode);
+
   /**
    * Fire accessible event of the given type for the given accessible.
    */
