@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=2 sw=2 et tw=78: */
 /* ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
  *
@@ -50,19 +51,19 @@
 #include "nsIDOMDocument.h"
 #include "nsIDOMAttr.h"
 #include "nsIDocument.h"
-#include "nsIDOMEventReceiver.h" 
+#include "nsIDOMEventTarget.h" 
 #include "nsIDOMNSEvent.h"
 #include "nsIDOMKeyEvent.h"
 #include "nsIDOMKeyListener.h" 
 #include "nsIDOMMouseListener.h"
 #include "nsIDOMMouseEvent.h"
+#include "nsIDOMComment.h"
 #include "nsISelection.h"
 #include "nsISelectionPrivate.h"
 #include "nsIDOMHTMLAnchorElement.h"
 #include "nsIDOMHTMLImageElement.h"
 #include "nsISelectionController.h"
 #include "nsIFileChannel.h"
-#include "nsIFrameSelection.h"  // For TABLESELECTION_ defines
 
 #include "nsICSSLoader.h"
 #include "nsICSSStyleSheet.h"
@@ -74,7 +75,6 @@
 #include "nsIContentIterator.h"
 #include "nsIDOMRange.h"
 #include "nsIDOMNSRange.h"
-#include "nsISupportsArray.h"
 #include "nsCOMArray.h"
 #include "nsVoidArray.h"
 #include "nsIFile.h"
@@ -133,6 +133,7 @@
 #include "nsIPrefBranch.h"
 #include "nsIPrefService.h"
 #include "nsIContentFilter.h"
+#include "nsEventDispatcher.h"
 
 const PRUnichar nbsp = 160;
 
@@ -141,11 +142,15 @@ static NS_DEFINE_CID(kCParserCID,     NS_PARSER_CID);
 // private clipboard data flavors for html copy/paste
 #define kHTMLContext   "text/_moz_htmlcontext"
 #define kHTMLInfo      "text/_moz_htmlinfo"
+#define kInsertCookie  "_moz_Insert Here_moz_"
+
+#define NS_FOUND_TARGET NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_EDITOR, 3)
 
 // some little helpers
 static PRInt32 FindPositiveIntegerAfterString(const char *aLeadingString, nsCString &aCStr);
 static nsresult RemoveFragComments(nsCString &theStr);
 static void RemoveBodyAndHead(nsIDOMNode *aNode);
+static nsresult FindTargetNode(nsIDOMNode *aStart, nsCOMPtr<nsIDOMNode> &aResult);
 
 static nsCOMPtr<nsIDOMNode> GetListParent(nsIDOMNode* aNode)
 {
@@ -281,31 +286,20 @@ nsHTMLEditor::InsertHTMLWithContext(const nsAString & aInputString,
   if (NS_FAILED(res)) return res;
   
   // create a dom document fragment that represents the structure to paste
-  nsCOMPtr<nsIDOMNode> fragmentAsNode;
-  PRInt32 rangeStartHint, rangeEndHint;
+  nsCOMPtr<nsIDOMNode> fragmentAsNode, streamStartParent, streamEndParent;
+  PRInt32 streamStartOffset = 0, streamEndOffset = 0;
 
-  nsAutoString contextStr;
-  contextStr.Assign(aContextStr);
-
-#ifdef MOZ_THUNDERBIRD
-  // See Bug #228920 --> editor / parser has trouble inserting single cell data from Excel.
-  // The details are in the bug. Until we figure out why the parser is not building the right 
-  // document structure for the single cell paste case, we can explicitly check for just such  
-  // a condition and work around it. By setting the contextStr to an empty string we end up 
-  // pasting just the cell text which is what we want anyway. 
-  // A paste from an excel cell always starts with a new line, two spaces and then the td tag
-  if (StringBeginsWith(aInputString, NS_LITERAL_STRING("\n  <td"))) 
-    contextStr.Truncate();
-#endif
-
-  res = CreateDOMFragmentFromPaste(aInputString, contextStr, aInfoStr, 
-                                            address_of(fragmentAsNode),
-                                            &rangeStartHint, &rangeEndHint);
+  res = CreateDOMFragmentFromPaste(aInputString, aContextStr, aInfoStr, 
+                                   address_of(fragmentAsNode),
+                                   address_of(streamStartParent),
+                                   address_of(streamEndParent),
+                                   &streamStartOffset,
+                                   &streamEndOffset);
   NS_ENSURE_SUCCESS(res, res);
 
-  nsCOMPtr<nsIDOMNode> targetNode, streamStartParent, streamEndParent, tempNode;
-  PRInt32 targetOffset=0, streamStartOffset=0, streamEndOffset=0, k;
-  
+  nsCOMPtr<nsIDOMNode> targetNode, tempNode;
+  PRInt32 targetOffset=0;
+
   if (!aDestNode)
   {
     // if caller didn't provide the destination/target node,
@@ -319,30 +313,6 @@ nsHTMLEditor::InsertHTMLWithContext(const nsAString & aInputString,
     targetNode = aDestNode;
     targetOffset = aDestOffset;
   }
-  
-  // fetch the start parent/offset by walking down the leading edge of 
-  // the fragmentAsNode tree (rangeStartHint # of times)
-  streamStartParent = fragmentAsNode;
-  for (k = 0; k < rangeStartHint; k++)
-  {
-    streamStartParent->GetFirstChild(getter_AddRefs(tempNode));
-    if (!tempNode) return NS_ERROR_FAILURE;
-    streamStartParent = tempNode;
-  }
-  // streamStartOffset is just always zero at this point
-  
-  // fetch the end parent/offset by walking down the trailing edge of 
-  // the fragmentAsNode tree (rangeEndHint # of times)
-  streamEndParent = fragmentAsNode;
-  for (k = 0; k < rangeEndHint; k++)
-  {
-    streamEndParent->GetLastChild(getter_AddRefs(tempNode));
-    if (!tempNode) return NS_ERROR_FAILURE;
-    streamEndParent = tempNode;
-  }
-  // streamEndOffset is just always after last child of streamEndParent at this point
-  res = GetLengthOfDOMNode(streamEndParent, (PRUint32&)streamEndOffset);
-  if (NS_FAILED(res)) return res;
 
   PRBool doContinue = PR_TRUE;
 
@@ -687,7 +657,7 @@ nsHTMLEditor::InsertHTMLWithContext(const nsAString & aInputString,
       nsCOMPtr<nsIDOMNode> selNode, tmp, visNode, highTable;
       PRInt32 selOffset;
       
-      // but dont cross tables
+      // but don't cross tables
       if (!nsHTMLEditUtils::IsTable(lastInsertNode))
       {
         res = GetLastEditableLeaf(lastInsertNode, address_of(selNode));
@@ -719,7 +689,7 @@ nsHTMLEditor::InsertHTMLWithContext(const nsAString & aInputString,
         if (NS_FAILED(res)) return res;
       }
       
-      // make sure we dont end up with selection collapsed after an invisible break node
+      // make sure we don't end up with selection collapsed after an invisible break node
       nsWSRunObject wsRunObj(this, selNode, selOffset);
       PRInt32 outVisOffset=0;
       PRInt16 visType=0;
@@ -733,7 +703,7 @@ nsHTMLEditor::InsertHTMLWithContext(const nsAString & aInputString,
         // out of the nsWSRunObject's state.
         if (!IsVisBreak(wsRunObj.mStartReasonNode))
         {
-          // dont leave selection past an invisible break;
+          // don't leave selection past an invisible break;
           // reset {selNode,selOffset} to point before break
           res = GetNodeLocation(wsRunObj.mStartReasonNode, address_of(selNode), &selOffset);
           // we want to be inside any inline style prior to break
@@ -938,7 +908,7 @@ nsHTMLEditor::RelativizeURIForNode(nsIDOMNode *aNode, nsIURL *aDestURL)
     attrNode->GetNodeValue(oldValue);
     if (!oldValue.IsEmpty())
     {
-      NS_ConvertUCS2toUTF8 oldCValue(oldValue);
+      NS_ConvertUTF16toUTF8 oldCValue(oldValue);
       nsCOMPtr<nsIURI> currentNodeURI;
       rv = NS_NewURI(getter_AddRefs(currentNodeURI), oldCValue);
       if (NS_SUCCEEDED(rv))
@@ -947,7 +917,7 @@ nsHTMLEditor::RelativizeURIForNode(nsIDOMNode *aNode, nsIURL *aDestURL)
         aDestURL->GetRelativeSpec(currentNodeURI, newRelativePath);
         if (!newRelativePath.IsEmpty())
         {
-          NS_ConvertUTF8toUCS2 newCValue(newRelativePath);
+          NS_ConvertUTF8toUTF16 newCValue(newRelativePath);
           attrNode->SetNodeValue(newCValue);
         }
       }
@@ -958,7 +928,7 @@ nsHTMLEditor::RelativizeURIForNode(nsIDOMNode *aNode, nsIURL *aDestURL)
 }
 
 nsresult
-nsHTMLEditor::RelativizeURIInFragmentList(nsCOMArray<nsIDOMNode> aNodeList,
+nsHTMLEditor::RelativizeURIInFragmentList(const nsCOMArray<nsIDOMNode> &aNodeList,
                                           const nsAString &aFlavor,
                                           nsIDOMDocument *aSourceDoc,
                                           nsIDOMNode *aTargetNode)
@@ -1092,7 +1062,7 @@ nsHTMLEditor::StripFormattingNodes(nsIDOMNode *aNode, PRBool aListOnly)
 
   nsresult res = NS_OK;
   nsCOMPtr<nsIContent> content = do_QueryInterface(aNode);
-  if (IsEmptyTextContent(content))
+  if (content->TextIsOnlyWhitespace())
   {
     nsCOMPtr<nsIDOMNode> parent, ignored;
     aNode->GetParentNode(getter_AddRefs(parent));
@@ -1147,10 +1117,8 @@ NS_IMETHODIMP nsHTMLEditor::PrepareHTMLTransferable(nsITransferable **aTransfera
       }
       (*aTransferable)->AddDataFlavor(kHTMLMime);
       (*aTransferable)->AddDataFlavor(kFileMime);
-#ifdef XP_WIN32
-      // image pasting from the clipboard is only implemented on Windows right now.
+      // image pasting from the clipboard is only implemented on Windows & Mac right now.
       (*aTransferable)->AddDataFlavor(kJPEGImageMime);
-#endif
     }
     (*aTransferable)->AddDataFlavor(kUnicodeMime);
   }
@@ -1211,8 +1179,9 @@ nsHTMLEditor::ParseCFHTML(nsCString & aCfhtml, PRUnichar **aStuffToPaste, PRUnic
  
   // create context string
   nsCAutoString contextUTF8(Substring(aCfhtml, startHTML, startFragment - startHTML) +
+                            NS_LITERAL_CSTRING("<!--" kInsertCookie "-->") +
                             Substring(aCfhtml, endFragment, endHTML - endFragment));
-  
+
   // validate startFragment
   // make sure it's not in the middle of a HTML tag
   // see bug #228879 for more details
@@ -1253,8 +1222,8 @@ nsHTMLEditor::ParseCFHTML(nsCString & aCfhtml, PRUnichar **aStuffToPaste, PRUnic
   RemoveFragComments(contextUTF8);
 
   // convert both strings to usc2
-  const nsAFlatString& fragUcs2Str = NS_ConvertUTF8toUCS2(fragmentUTF8);
-  const nsAFlatString& cntxtUcs2Str = NS_ConvertUTF8toUCS2(contextUTF8);
+  const nsAFlatString& fragUcs2Str = NS_ConvertUTF8toUTF16(fragmentUTF8);
+  const nsAFlatString& cntxtUcs2Str = NS_ConvertUTF8toUTF16(contextUTF8);
   
   // translate platform linebreaks for fragment
   PRInt32 oldLengthInChars = fragUcs2Str.Length() + 1;  // +1 to include null terminator
@@ -1470,6 +1439,9 @@ NS_IMETHODIMP nsHTMLEditor::InsertFromTransferable(nsITransferable *transferable
   }
       
   // Try to scroll the selection into view if the paste/drop succeeded
+
+  // After ScrollSelectionIntoView(), the pending notifications might be
+  // flushed and PresShell/PresContext/Frames may be dead. See bug 418470.
   if (NS_SUCCEEDED(rv))
     ScrollSelectionIntoView(PR_FALSE);
 
@@ -1652,11 +1624,11 @@ NS_IMETHODIMP nsHTMLEditor::InsertFromDrop(nsIDOMEvent* aDropEvent)
 
           rv = selection->GetRangeAt(j, getter_AddRefs(range));
           if (NS_FAILED(rv) || !range) 
-            continue;//dont bail yet, iterate through them all
+            continue;//don't bail yet, iterate through them all
 
           nsCOMPtr<nsIDOMNSRange> nsrange(do_QueryInterface(range));
           if (NS_FAILED(rv) || !nsrange) 
-            continue;//dont bail yet, iterate through them all
+            continue;//don't bail yet, iterate through them all
 
           rv = nsrange->IsPointInRange(newSelectionParent, newSelectionOffset, &cursorIsInSelection);
           if(cursorIsInSelection)
@@ -1698,6 +1670,8 @@ NS_IMETHODIMP nsHTMLEditor::InsertFromDrop(nsIDOMEvent* aDropEvent)
     if (!nsEditorHookUtils::DoInsertionHook(domdoc, aDropEvent, trans))
       return NS_OK;
 
+    // Beware! This may flush notifications via synchronous
+    // ScrollSelectionIntoView.
     rv = InsertFromTransferable(trans, srcdomdoc, contextStr, infoStr,
                                 newSelectionParent,
                                 newSelectionOffset, deleteSelection);
@@ -1733,7 +1707,7 @@ nsHTMLEditor::PutDragDataInTransferable(nsITransferable **aTransferable)
   if (!bIsPlainTextControl)
   {
     // encode the selection as html with contextual info
-    rv = docEncoder->EncodeToStringWithContext(buffer, parents, info);
+    rv = docEncoder->EncodeToStringWithContext(parents, info, buffer);
     if (NS_FAILED(rv)) return rv;
   }
   else
@@ -1831,19 +1805,12 @@ PRBool nsHTMLEditor::HavePrivateHTMLFlavor(nsIClipboard *aClipboard)
   
   if (!aClipboard) return PR_FALSE;
   PRBool bHavePrivateHTMLFlavor = PR_FALSE;
-  nsCOMPtr<nsISupportsArray> flavArray;
   
-  nsresult res = NS_NewISupportsArray(getter_AddRefs(flavArray));
-  if (NS_FAILED(res)) return PR_FALSE;
+  const char* flavArray[] = { kHTMLContext };
   
-  nsCOMPtr<nsISupportsCString> contextString = do_CreateInstance(NS_SUPPORTS_CSTRING_CONTRACTID);
-  if (!contextString) return PR_FALSE;
-  
-  contextString->SetData(NS_LITERAL_CSTRING(kHTMLContext));
-  
-  flavArray->AppendElement(contextString);
-  
-  if (NS_SUCCEEDED(aClipboard->HasDataMatchingFlavors (flavArray, nsIClipboard::kGlobalClipboard, &bHavePrivateHTMLFlavor )))
+  if (NS_SUCCEEDED(aClipboard->HasDataMatchingFlavors(flavArray,
+    NS_ARRAY_LENGTH(flavArray), nsIClipboard::kGlobalClipboard,
+    &bHavePrivateHTMLFlavor )))
     return bHavePrivateHTMLFlavor;
     
   return PR_FALSE;
@@ -1854,8 +1821,12 @@ NS_IMETHODIMP nsHTMLEditor::Paste(PRInt32 aSelectionType)
 {
   ForceCompositionEnd();
 
+  PRBool preventDefault;
+  nsresult rv = FireClipboardEvent(NS_PASTE, &preventDefault);
+  if (NS_FAILED(rv) || preventDefault)
+    return rv;
+
   // Get Clipboard Service
-  nsresult rv;
   nsCOMPtr<nsIClipboard> clipboard(do_GetService("@mozilla.org/widget/clipboard;1", &rv));
   if (NS_FAILED(rv))
     return rv;
@@ -1921,6 +1892,8 @@ NS_IMETHODIMP nsHTMLEditor::Paste(PRInt32 aSelectionType)
       if (!nsEditorHookUtils::DoInsertionHook(domdoc, nsnull, trans))
         return NS_OK;
 
+      // Beware! This may flush notifications via synchronous
+      // ScrollSelectionIntoView.
       rv = InsertFromTransferable(trans, nsnull, contextStr, infoStr,
                                   nsnull, 0, PR_TRUE);
     }
@@ -1952,6 +1925,8 @@ NS_IMETHODIMP nsHTMLEditor::PasteNoFormatting(PRInt32 aSelectionType)
     if (NS_SUCCEEDED(clipboard->GetData(trans, aSelectionType)) && IsModifiable())
     {
       const nsAFlatString& empty = EmptyString();
+      // Beware! This may flush notifications via synchronous
+      // ScrollSelectionIntoView.
       rv = InsertFromTransferable(trans, nsnull, empty, empty, nsnull, 0,
                                   PR_TRUE);
     }
@@ -1963,58 +1938,37 @@ NS_IMETHODIMP nsHTMLEditor::PasteNoFormatting(PRInt32 aSelectionType)
 
 NS_IMETHODIMP nsHTMLEditor::CanPaste(PRInt32 aSelectionType, PRBool *aCanPaste)
 {
-  if (!aCanPaste)
-    return NS_ERROR_NULL_POINTER;
+  NS_ENSURE_ARG_POINTER(aCanPaste);
   *aCanPaste = PR_FALSE;
-  
+
   // can't paste if readonly
   if (!IsModifiable())
     return NS_OK;
-    
+
   nsresult rv;
   nsCOMPtr<nsIClipboard> clipboard(do_GetService("@mozilla.org/widget/clipboard;1", &rv));
   if (NS_FAILED(rv)) return rv;
   
   // the flavors that we can deal with
-  const char* const textEditorFlavors[] = { kUnicodeMime, nsnull };
-  const char* const htmlEditorFlavors[] = { kHTMLMime, kJPEGImageMime, nsnull };
+  const char* textEditorFlavors[] = { kUnicodeMime };
+  const char* textHtmlEditorFlavors[] = { kUnicodeMime, kHTMLMime,
+                                          kJPEGImageMime };
 
-  nsCOMPtr<nsISupportsArray> flavorsList =
-                           do_CreateInstance(NS_SUPPORTSARRAY_CONTRACTID, &rv);
-  if (NS_FAILED(rv)) return rv;
-  
   PRUint32 editorFlags;
   GetFlags(&editorFlags);
   
-  // add the flavors for all editors
-  for (const char* const* flavor = textEditorFlavors; *flavor; flavor++)
-  {
-    nsCOMPtr<nsISupportsCString> flavorString  = do_CreateInstance(NS_SUPPORTS_CSTRING_CONTRACTID);
-    if (flavorString)
-    {
-      flavorString->SetData(nsDependentCString(*flavor));
-      flavorsList->AppendElement(flavorString);
-    }
-  }
-  
-  // add the HTML-editor only flavors
-  if ((editorFlags & eEditorPlaintextMask) == 0)
-  {
-    for (const char* const* htmlFlavor = htmlEditorFlavors;
-         *htmlFlavor;
-         htmlFlavor++)
-    {
-      nsCOMPtr<nsISupportsCString> flavorString  = do_CreateInstance(NS_SUPPORTS_CSTRING_CONTRACTID);
-      if (flavorString)
-      {
-        flavorString->SetData(nsDependentCString(*htmlFlavor));
-        flavorsList->AppendElement(flavorString);
-      }
-    }
-  }
-  
   PRBool haveFlavors;
-  rv = clipboard->HasDataMatchingFlavors(flavorsList, aSelectionType, &haveFlavors);
+
+  // Use the flavors depending on the current editor mask
+  if ((editorFlags & eEditorPlaintextMask))
+    rv = clipboard->HasDataMatchingFlavors(textEditorFlavors,
+                                           NS_ARRAY_LENGTH(textEditorFlavors),
+                                           aSelectionType, &haveFlavors);
+  else
+    rv = clipboard->HasDataMatchingFlavors(textHtmlEditorFlavors,
+                                           NS_ARRAY_LENGTH(textHtmlEditorFlavors),
+                                           aSelectionType, &haveFlavors);
+  
   if (NS_FAILED(rv)) return rv;
   
   *aCanPaste = haveFlavors;
@@ -2210,7 +2164,7 @@ nsHTMLEditor::InsertTextWithQuotations(const nsAString &aStringToInsert)
 #ifdef DEBUG_akkana_verbose
     printf("==== Inserting text as %squoted: ---\n%s---\n",
            curHunkIsQuoted ? "" : "non-",
-           NS_LossyConvertUCS2toASCII(curHunk).get());
+           NS_LossyConvertUTF16toASCII(curHunk).get());
 #endif
     if (curHunkIsQuoted)
       rv = InsertAsPlaintextQuotation(curHunk, PR_FALSE,
@@ -2257,7 +2211,7 @@ nsHTMLEditor::InsertAsPlaintextQuotation(const nsAString & aQuotedText,
 
   // The quotesPreformatted pref is a temporary measure. See bug 69638.
   // Eventually we'll pick one way or the other.
-  PRBool quotesInPre;
+  PRBool quotesInPre = PR_FALSE;
   nsCOMPtr<nsIPrefBranch> prefBranch =
     do_GetService(NS_PREFSERVICE_CONTRACTID, &rv);
   if (NS_SUCCEEDED(rv) && prefBranch)
@@ -2480,15 +2434,84 @@ void RemoveBodyAndHead(nsIDOMNode *aNode)
   }
 }
 
+/**
+ * This function finds the target node that we will be pasting into. aStart is
+ * the context that we're given and aResult will be the target. Initially,
+ * *aResult must be NULL.
+ *
+ * The target for a paste is found by either finding the node that contains
+ * the magical comment node containing kInsertCookie or, failing that, the
+ * firstChild of the firstChild (until we reach a leaf).
+ */
+nsresult FindTargetNode(nsIDOMNode *aStart, nsCOMPtr<nsIDOMNode> &aResult)
+{
+  if (!aStart)
+    return NS_OK;
+
+  nsCOMPtr<nsIDOMNode> child, tmp;
+
+  nsresult rv = aStart->GetFirstChild(getter_AddRefs(child));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  if (!child)
+  {
+    // If the current result is NULL, then aStart is a leaf, and is the
+    // fallback result.
+    if (!aResult)
+      aResult = aStart;
+
+    return NS_OK;
+  }
+
+  do
+  {
+    // Is this child the magical cookie?
+    nsCOMPtr<nsIDOMComment> comment = do_QueryInterface(child);
+    if (comment)
+    {
+      nsAutoString data;
+      rv = comment->GetData(data);
+      NS_ENSURE_SUCCESS(rv, rv);
+
+      if (data.EqualsLiteral(kInsertCookie))
+      {
+        // Yes it is! Return an error so we bubble out and short-circuit the
+        // search.
+        aResult = aStart;
+
+        // Note: it doesn't matter if this fails.
+        aStart->RemoveChild(child, getter_AddRefs(tmp));
+
+        return NS_FOUND_TARGET;
+      }
+    }
+
+    // Note: Don't use NS_ENSURE_* here since we return a failure result to
+    // inicate that we found the magical cookie and we don't want to spam the
+    // console.
+    rv = FindTargetNode(child, aResult);
+    if (NS_FAILED(rv))
+      return rv;
+
+    rv = child->GetNextSibling(getter_AddRefs(tmp));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    child = tmp;
+  } while (child);
+
+  return NS_OK;
+}
 
 nsresult nsHTMLEditor::CreateDOMFragmentFromPaste(const nsAString &aInputString,
                                                   const nsAString & aContextStr,
                                                   const nsAString & aInfoStr,
                                                   nsCOMPtr<nsIDOMNode> *outFragNode,
-                                                  PRInt32 *outRangeStartHint,
-                                                  PRInt32 *outRangeEndHint)
+                                                  nsCOMPtr<nsIDOMNode> *outStartNode,
+                                                  nsCOMPtr<nsIDOMNode> *outEndNode,
+                                                  PRInt32 *outStartOffset,
+                                                  PRInt32 *outEndOffset)
 {
-  if (!outFragNode || !outRangeStartHint || !outRangeEndHint) 
+  if (!outFragNode || !outStartNode || !outEndNode) 
     return NS_ERROR_NULL_POINTER;
   nsCOMPtr<nsIDOMDocumentFragment> docfrag;
   nsCOMPtr<nsIDOMNode> contextAsNode, tmp;  
@@ -2501,10 +2524,9 @@ nsresult nsHTMLEditor::CreateDOMFragmentFromPaste(const nsAString &aInputString,
   NS_ENSURE_TRUE(doc, NS_ERROR_FAILURE);
   
   // if we have context info, create a fragment for that
-  nsVoidArray tagStack;
+  nsAutoTArray<nsString, 32> tagStack;
   nsCOMPtr<nsIDOMDocumentFragment> contextfrag;
   nsCOMPtr<nsIDOMNode> contextLeaf, junk;
-  PRInt32 contextDepth = 0;
   if (!aContextStr.IsEmpty())
   {
     res = ParseFragment(aContextStr, tagStack, doc, address_of(contextAsNode));
@@ -2513,73 +2535,86 @@ nsresult nsHTMLEditor::CreateDOMFragmentFromPaste(const nsAString &aInputString,
 
     res = StripFormattingNodes(contextAsNode);
     NS_ENSURE_SUCCESS(res, res);
-    
+
     RemoveBodyAndHead(contextAsNode);
-    
-    // cache the deepest leaf in the context
-    tmp = contextAsNode;
-    while (tmp)
-    {
-      contextDepth++;
-      contextLeaf = tmp;
-      contextLeaf->GetFirstChild(getter_AddRefs(tmp));
-    }
+
+    res = FindTargetNode(contextAsNode, contextLeaf);
+    if (res == NS_FOUND_TARGET)
+      res = NS_OK;
+    NS_ENSURE_SUCCESS(res, res);
   }
- 
+
   // get the tagstack for the context
   res = CreateTagStack(tagStack, contextLeaf);
-  if (NS_FAILED(res))
-  {
-    FreeTagStackStrings(tagStack);
-    return res;
-  }
+  NS_ENSURE_SUCCESS(res, res);
+
   // create fragment for pasted html
   res = ParseFragment(aInputString, tagStack, doc, outFragNode);
-  FreeTagStackStrings(tagStack);
   NS_ENSURE_SUCCESS(res, res);
   NS_ENSURE_TRUE(*outFragNode, NS_ERROR_FAILURE);
-      
+
   RemoveBodyAndHead(*outFragNode);
-  
+
   if (contextAsNode)
   {
     // unite the two trees
     contextLeaf->AppendChild(*outFragNode, getter_AddRefs(junk));
     *outFragNode = contextAsNode;
-    // no longer have fragmentAsNode in tree
-    contextDepth--;
   }
- 
+
   res = StripFormattingNodes(*outFragNode, PR_TRUE);
   NS_ENSURE_SUCCESS(res, res);
- 
+
+  // If there was no context, then treat all of the data we did get as the
+  // pasted data.
+  if (contextLeaf)
+    *outEndNode = *outStartNode = contextLeaf;
+  else
+    *outEndNode = *outStartNode = *outFragNode;
+
+  *outStartOffset = 0;
+
   // get the infoString contents
   nsAutoString numstr1, numstr2;
   if (!aInfoStr.IsEmpty())
   {
-    PRInt32 err, sep;
+    PRInt32 err, sep, num;
     sep = aInfoStr.FindChar((PRUnichar)',');
     numstr1 = Substring(aInfoStr, 0, sep);
     numstr2 = Substring(aInfoStr, sep+1, aInfoStr.Length() - (sep+1));
-    *outRangeStartHint = numstr1.ToInteger(&err) + contextDepth;
-    *outRangeEndHint   = numstr2.ToInteger(&err) + contextDepth;
-  }
-  else
-  {
-    *outRangeStartHint = contextDepth;
-    *outRangeEndHint = contextDepth;
+
+    // Move the start and end children.
+    num = numstr1.ToInteger(&err);
+    while (num--)
+    {
+      (*outStartNode)->GetFirstChild(getter_AddRefs(tmp));
+      if (!tmp)
+        return NS_ERROR_FAILURE;
+      tmp.swap(*outStartNode);
+    }
+
+    num = numstr2.ToInteger(&err);
+    while (num--)
+    {
+      (*outEndNode)->GetLastChild(getter_AddRefs(tmp));
+      if (!tmp)
+        return NS_ERROR_FAILURE;
+      tmp.swap(*outEndNode);
+    }
   }
+
+  GetLengthOfDOMNode(*outEndNode, (PRUint32&)*outEndOffset);
   return res;
 }
 
 
 nsresult nsHTMLEditor::ParseFragment(const nsAString & aFragStr,
-                                     nsVoidArray &aTagStack,
+                                     nsTArray<nsString> &aTagStack,
                                      nsIDocument* aTargetDocument,
                                      nsCOMPtr<nsIDOMNode> *outNode)
 {
   // figure out if we are parsing full context or not
-  PRBool bContext = (aTagStack.Count()==0);
+  PRBool bContext = aTagStack.IsEmpty();
 
   // create the parser to do the conversion.
   nsresult res;
@@ -2599,23 +2634,23 @@ nsresult nsHTMLEditor::ParseFragment(const nsAString & aFragStr,
   NS_ENSURE_TRUE(fragSink, NS_ERROR_FAILURE);
 
   fragSink->SetTargetDocument(aTargetDocument);
-  
+
   // parse the fragment
   parser->SetContentSink(sink);
   if (bContext)
-    parser->Parse(aFragStr, (void*)0, NS_LITERAL_CSTRING("text/html"), PR_FALSE, PR_TRUE, eDTDMode_fragment);
+    parser->Parse(aFragStr, (void*)0, NS_LITERAL_CSTRING("text/html"), PR_TRUE, eDTDMode_fragment);
   else
     parser->ParseFragment(aFragStr, 0, aTagStack, PR_FALSE, NS_LITERAL_CSTRING("text/html"), eDTDMode_quirks);
   // get the fragment node
   nsCOMPtr<nsIDOMDocumentFragment> contextfrag;
-  res = fragSink->GetFragment(getter_AddRefs(contextfrag));
+  res = fragSink->GetFragment(PR_TRUE, getter_AddRefs(contextfrag));
   NS_ENSURE_SUCCESS(res, res);
   *outNode = do_QueryInterface(contextfrag);
   
   return res;
 }
 
-nsresult nsHTMLEditor::CreateTagStack(nsVoidArray &aTagStack, nsIDOMNode *aNode)
+nsresult nsHTMLEditor::CreateTagStack(nsTArray<nsString> &aTagStack, nsIDOMNode *aNode)
 {
   nsresult res = NS_OK;
   nsCOMPtr<nsIDOMNode> node= aNode;
@@ -2631,15 +2666,11 @@ nsresult nsHTMLEditor::CreateTagStack(nsVoidArray &aTagStack, nsIDOMNode *aNode)
     node->GetNodeType(&nodeType);
     if (nsIDOMNode::ELEMENT_NODE == nodeType)
     {
-      nsAutoString tagName;
-      node->GetNodeName(tagName);
-      // XXX Wish we didn't have to allocate here
-      PRUnichar* name = ToNewUnicode(tagName);
-      if (!name) 
-	      return NS_ERROR_OUT_OF_MEMORY;
+      nsString* tagName = aTagStack.AppendElement();
+      NS_ENSURE_TRUE(tagName, NS_ERROR_OUT_OF_MEMORY);
 
-      aTagStack.AppendElement(name);
-      // printf("%s\n",NS_LossyConvertUCS2toASCII(tagName).get());
+      node->GetNodeName(*tagName);
+      // printf("%s\n",NS_LossyConvertUTF16toASCII(tagName).get());
     }
 
     res = temp->GetParentNode(getter_AddRefs(node));
@@ -2648,25 +2679,12 @@ nsresult nsHTMLEditor::CreateTagStack(nsVoidArray &aTagStack, nsIDOMNode *aNode)
   
   if (!bSeenBody)
   {
-      PRUnichar* bodyname = ToNewUnicode(NS_LITERAL_STRING("BODY"));
-      aTagStack.AppendElement(bodyname);
+      aTagStack.AppendElement(NS_LITERAL_STRING("BODY"));
   }
   return res;
 }
 
 
-void nsHTMLEditor::FreeTagStackStrings(nsVoidArray &tagStack)
-{
-  PRInt32 count = tagStack.Count();
-  for (PRInt32 i = 0; i < count; i++) 
-  {
-    PRUnichar* str = (PRUnichar*)tagStack.ElementAt(i);
-    if (str) {
-      NS_Free(str);
-    }
-  }
-}
-
 nsresult nsHTMLEditor::CreateListOfNodesToPaste(nsIDOMNode  *aFragmentAsNode,
                                                 nsCOMArray<nsIDOMNode>& outNodeList,
                                                 nsIDOMNode *aStartNode,
@@ -2861,7 +2879,7 @@ nsHTMLEditor::ReplaceOrphanedStructure(PRBool aEnd,
   if (replaceNode)
   {
     // postprocess list to remove any descendants of this node
-    // so that we dont insert them twice.
+    // so that we don't insert them twice.
     nsCOMPtr<nsIDOMNode> endpoint;
     do
     {
