@@ -97,17 +97,6 @@
 
 static const char *sJSStackContractID="@mozilla.org/js/xpc/ContextStack;1";
 
-//
-// This is an attempt to fix bug 20573.
-//
-#ifdef XP_UNIX
-static const char *AppendXGeometryNameClass = "gnuzilla.geometry";
-static       char *AppendXGeometryString = NULL; // uses lazy initialization
-static       char  AppendXGeometryBuffer[240];
-extern int    gArgc;
-extern char **gArgv;
-#endif
-
 /****************************************************************
  ******************** nsWatcherWindowEntry **********************
  ****************************************************************/
@@ -471,175 +460,6 @@ struct SizeSpec {
   }
 };
 
-#if XP_UNIX | HAVE_XRM
-#include <X11/Xlib.h>
-#include <X11/Xresource.h>
-#include <X11/Xutil.h>
-//
-// This is an attempt to fix bug 20573.
-//
-// Unfortunately, a position value of +0+0 causes something later in the
-// processing to fail to set the position.
-//
-// Negative position numbers might not work.
-//
-void AppendXRDBGeometryFeatures( const char    *aName,
-                                 PRBool        featuresSpecified,
-                                 nsCAutoString *featuresp
-                               ) {
-
-  if (NULL != aName) { // name is not null
-    if (0 == strcmp( aName, "_blank")) { // name is _blank
-      if (NULL == AppendXGeometryString) { // build AppendXGeometryString
-        //
-        // This runs only once.  Before it runs, AppendXGeometryString is
-        // NULL.  After it runs, AppendXGeometryString points to the empty
-        // string or to a string of features to be appended.
-        //
-        char *buf = AppendXGeometryBuffer; // shorthand
-        buf[0] = '\0'; // initialize to the empty string
-        AppendXGeometryString = buf; // leave it as empty string or filled
-
-
-        char *theaddr = NULL;
-
-        for (int i = 0; i < gArgc - 1; i++){
-          if(!strcmp(gArgv[i], "-geometry")){
-            theaddr = gArgv[i + 1];
-            break;
-          }
-        }
-
-        if(!theaddr)
-        {
-          //
-          Display *thedisp = XOpenDisplay("");
-          if (NULL == thedisp) {
-            //
-            // Could not open the display.
-            // Should issue a warning or error message.
-            //fprintf( stderr, "*** XOpenDisplay failed.\n");
-            //
-            return;
-          }
-          char *therm = XResourceManagerString(thedisp);
-          if (NULL == therm) {
-            //
-            // Could not get the resource manager string.
-            // Should issue a warning or error message.
-            //fprintf( stderr, "*** XResourceManagerString failed.\n");
-            //
-            return;
-          }
-          XrmInitialize();
-          XrmDatabase thedb = XrmGetStringDatabase(therm);
-          if (NULL == thedb) {
-            //
-            // Could not create a database.
-            // Should issue a warning or error message.
-            //fprintf( stderr, "*** XrmGetStringDatabase failed.\n");
-            //
-            return;
-          }
-          char *type = "badstuff";
-          XrmValue val;
-          const char *nameclass = AppendXGeometryNameClass; // shorthand
-          Bool gotrs = XrmGetResource( thedb, nameclass, nameclass, &type, &val);
-          if (False == gotrs) {
-            //
-            // Did not find a usable resource entry.
-            //
-            // Leave AppendXGeometryString pointing to the empty string.
-            //
-            // Do not issue a warning or error message.
-            //fprintf( stderr, "*** XrmGetResource failed.\n");
-            //
-            return;
-          }
-          if (0 != strcmp( "String", type)) {
-            //
-            // Geometry entry was of the wrong type.
-            // Should issue a warning or error message.
-            //fprintf( stderr, "*** String type comparison failed.\n");
-            //
-            return;
-          }
-          int thesize = val.size;
-          theaddr = val.addr;
-          int thelen = strlen(theaddr);
-          if (1 != (thesize - thelen)) {
-            //
-            // The size and string length don't match.
-            // Should issue a warning or error message.
-            //fprintf( stderr, "*** size vs. len check failed.\n");
-            //
-            return;
-          }
-          XrmDestroyDatabase(thedb);
-        }
-
-        int x = 9999999;
-        int y = 9999999;
-        unsigned int wid = 9999999;
-        unsigned int hit = 9999999;
-
-        int gotwhat = XParseGeometry( theaddr, &x, &y, &wid, &hit);
-
-        if (0 != (gotwhat & XValue)) {
-          strncat( buf, "screenX=", 15);
-          if ((0 == x) && (0 != (gotwhat & XNegative))) { // handle minus zero
-            strncat( buf, "-", 3);
-          } // handle minus zero
-          snprintf( buf + strlen(buf), 20, "%d", x);
-        }
-        if (0 != (gotwhat & YValue)) {
-          strncat( buf, ",screenY=", 15);
-          if ((0 == y) && (0 != (gotwhat & YNegative))) { // handle minus zero
-            strncat( buf, "-", 3);
-          } // handle minus zero
-          snprintf( buf + strlen(buf), 20, "%d", y);
-        }
-        if (0 != (gotwhat & WidthValue)) {
-          strncat( buf, ",width=", 15);
-          snprintf( buf + strlen(buf), 20, "%d", wid);
-        }
-        if (0 != (gotwhat & HeightValue)) {
-          strncat( buf, ",height=", 15);
-          snprintf( buf + strlen(buf), 20, "%d", hit);
-        }
-        if (160 < strlen(buf)) {
-          //
-          // The string is longer than it should be.
-          // Should issue a warning or error message.
-          //// fprintf( stderr, "*** buffer too long.\n");
-          //
-          return;
-        }
-        //
-        // Don't free the resource manager string; Xlib owns it.
-        //
-        // For Firefox, don't close the display.
-        //
-        // AppendXGeometryString was already assigned to point at the buffer.
-        //
-      } // build AppendXGeometryString
-      if (0 < strlen(AppendXGeometryString)) { // have something to append
-        if (featuresSpecified) { // specified, must append
-          (*featuresp).Append(",");
-          (*featuresp).Append(AppendXGeometryString);
-          (*featuresp).StripWhitespace(); // for good measure
-        } // specified, must append
-        else { // not previously specified, must assign
-          (*featuresp).Assign(AppendXGeometryString);
-          featuresSpecified = PR_TRUE;
-          (*featuresp).StripWhitespace();
-        } // not previously specified, must assign
-      } // have something to append
-    } // name is _blank
-  } // name is not null
-}
-#endif
-
 NS_IMETHODIMP
 nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
                               const char *aUrl,
@@ -659,6 +479,7 @@ nsWindowWatcher::OpenWindowJS(nsIDOMWindow *aParent,
     if (argc == 0)
       argv = nsnull;
   }
+
   return OpenWindowJSInternal(aParent, aUrl, aName, aFeatures, aDialog,
                               argv, PR_TRUE, _retval);
 }
@@ -727,14 +548,6 @@ nsWindowWatcher::OpenWindowJSInternal(nsIDOMWindow *aParent,
 
   nsCOMPtr<nsIDOMChromeWindow> chromeParent(do_QueryInterface(aParent));
 
-  //
-  // This is an attempt to fix bug 20573.
-  //
-#if XP_UNIX | HAVE_XRM
-  if (aDialog == PR_FALSE)
-    AppendXRDBGeometryFeatures( aName, featuresSpecified, &features);
-#endif
-
   // Make sure we call CalculateChromeFlags() *before* we push the
   // callee context onto the context stack so that
   // CalculateChromeFlags() sees the actual caller when doing it's
