@@ -41,9 +41,6 @@
  * ***** END LICENSE BLOCK ***** */
 
 #include <ole2.h>
-#ifndef __MINGW32__
-#include <urlmon.h>
-#endif
 #include <shlobj.h>
 
 #include "nsDataObj.h"
@@ -68,6 +65,17 @@
 #include "nsCRT.h"
 #include "nsPrintfCString.h"
 #include "nsIStringBundle.h"
+#include "nsEscape.h"
+#include "nsIURL.h"
+#include "nsNetUtil.h"
+#include "nsXPCOMStrings.h"
+
+// XXX for older version of PSDK where IAsyncOperation and related stuff is not available
+// but this should be removed when parocles config is updated
+// IAsyncOperation interface GUID
+#ifndef __IAsyncOperation_INTERFACE_DEFINED__
+  const IID IID_IAsyncOperation = {0x3D8B0590, 0xF691, 0x11d2, {0x8E, 0xA9, 0x00, 0x60, 0x97, 0xDF, 0x5B, 0xD4}};
+#endif
 
 #if 0
 #define PRNTDEBUG(_x) printf(_x);
@@ -79,6 +87,258 @@
 #define PRNTDEBUG3(_x1, _x2, _x3) // printf(_x1, _x2, _x3);
 #endif
 
+//-----------------------------------------------------------------------------
+// CStream implementation
+nsDataObj::CStream::CStream() : mRefCount(1)
+{
+}
+
+//-----------------------------------------------------------------------------
+nsDataObj::CStream::~CStream()
+{
+}
+
+//-----------------------------------------------------------------------------
+// helper - initializes the stream
+nsresult nsDataObj::CStream::Init(nsIURI *pSourceURI)
+{
+  nsresult rv;
+  rv = NS_NewChannel(getter_AddRefs(mChannel), pSourceURI);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  rv = mChannel->Open(getter_AddRefs(mInputStream));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  return NS_OK;
+}
+
+//-----------------------------------------------------------------------------
+// IUnknown
+STDMETHODIMP nsDataObj::CStream::QueryInterface(REFIID refiid, void** ppvResult)
+{
+  *ppvResult = NULL;
+  if (IID_IUnknown == refiid ||
+      refiid == IID_IStream)
+
+  {
+    *ppvResult = this;
+  }
+
+  if (NULL != *ppvResult)
+  {
+    ((LPUNKNOWN)*ppvResult)->AddRef();
+    return S_OK;
+  }
+
+  return ResultFromScode(E_NOINTERFACE);
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP_(ULONG) nsDataObj::CStream::AddRef(void)
+{
+  return ++mRefCount;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP_(ULONG) nsDataObj::CStream::Release(void)
+{
+  ULONG nCount = --mRefCount;
+  if (nCount == 0)
+  {
+    delete this;
+    return (ULONG)0;
+  }
+
+  return mRefCount;
+}
+
+//-----------------------------------------------------------------------------
+// IStream
+STDMETHODIMP nsDataObj::CStream::Clone(IStream** ppStream)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Commit(DWORD dwFrags)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::CopyTo(IStream* pDestStream,
+                                        ULARGE_INTEGER nBytesToCopy,
+                                        ULARGE_INTEGER* nBytesRead,
+                                        ULARGE_INTEGER* nBytesWritten)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::LockRegion(ULARGE_INTEGER nStart,
+                                            ULARGE_INTEGER nBytes,
+                                            DWORD dwFlags)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Read(void* pvBuffer,
+                                      ULONG nBytesToRead,
+                                      ULONG* nBytesRead)
+{
+  NS_ENSURE_TRUE(mInputStream, E_FAIL);
+
+  nsresult rv;
+  PRUint32 read;
+  *nBytesRead = 0;
+
+  do {
+    read = 0;
+    rv = mInputStream->Read((char*)pvBuffer + *nBytesRead, nBytesToRead - *nBytesRead, &read);
+    NS_ENSURE_SUCCESS(rv, S_FALSE);
+
+    *nBytesRead += read;
+  } while ((*nBytesRead < nBytesToRead) && read);
+
+  return S_OK;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Revert(void)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Seek(LARGE_INTEGER nMove,
+                                      DWORD dwOrigin,
+                                      ULARGE_INTEGER* nNewPos)
+{
+  if (nNewPos == NULL)
+    return STG_E_INVALIDPOINTER;
+
+  if (nMove.LowPart == 0 && nMove.HighPart == 0 &&
+      (dwOrigin == STREAM_SEEK_SET || dwOrigin == STREAM_SEEK_CUR)) { 
+    nNewPos->LowPart = 0;
+    nNewPos->HighPart = 0;
+    return S_OK;
+  }
+
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::SetSize(ULARGE_INTEGER nNewSize)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Stat(STATSTG* statstg, DWORD dwFlags)
+{
+  if (statstg == NULL)
+    return STG_E_INVALIDPOINTER;
+
+  if (!mChannel)
+    return E_FAIL;
+
+  memset((void*)statstg, 0, sizeof(STATSTG));
+
+  if (dwFlags != STATFLAG_NONAME) 
+  {
+    nsCOMPtr<nsIURI> sourceURI;
+    if (NS_FAILED(mChannel->GetURI(getter_AddRefs(sourceURI)))) {
+      return E_FAIL;
+    }
+
+    nsCAutoString strFileName;
+    nsCOMPtr<nsIURL> sourceURL = do_QueryInterface(sourceURI);
+    sourceURL->GetFileName(strFileName);
+
+    if (strFileName.IsEmpty())
+      return E_FAIL;
+
+    NS_UnescapeURL(strFileName);
+    NS_ConvertUTF8toUTF16 wideFileName(strFileName);
+
+    PRUint32 nMaxNameLength = (wideFileName.Length()*2) + 2;
+    void * retBuf = CoTaskMemAlloc(nMaxNameLength); // freed by caller
+    if (!retBuf) 
+      return STG_E_INSUFFICIENTMEMORY;
+
+    ZeroMemory(retBuf, nMaxNameLength);
+    memcpy(retBuf, wideFileName.get(), wideFileName.Length()*2);
+    statstg->pwcsName = (LPOLESTR)retBuf;
+  }
+
+  SYSTEMTIME st;
+  FILETIME ft;
+
+  statstg->type = STGTY_STREAM;
+
+  GetSystemTime(&st);
+  SystemTimeToFileTime((const SYSTEMTIME*)&st, (LPFILETIME)&statstg->mtime);
+  statstg->ctime = statstg->atime = statstg->mtime;
+
+  PRInt32 nLength = 0;
+  if (mChannel)
+    mChannel->GetContentLength(&nLength);
+
+  if (nLength < 0) 
+    nLength = 0;
+
+  statstg->cbSize.LowPart = (DWORD)nLength;
+  statstg->grfMode = STGM_READ;
+  statstg->grfLocksSupported = LOCK_ONLYONCE;
+  statstg->clsid = CLSID_NULL;
+
+  return S_OK;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::UnlockRegion(ULARGE_INTEGER nStart,
+                                              ULARGE_INTEGER nBytes,
+                                              DWORD dwFlags)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+STDMETHODIMP nsDataObj::CStream::Write(const void* pvBuffer,
+                                       ULONG nBytesToRead,
+                                       ULONG* nBytesRead)
+{
+  return E_NOTIMPL;
+}
+
+//-----------------------------------------------------------------------------
+HRESULT nsDataObj::CreateStream(IStream **outStream)
+{
+  NS_ENSURE_TRUE(outStream, E_INVALIDARG);
+
+  nsresult rv = NS_ERROR_FAILURE;
+  nsAutoString wideFileName;
+  nsCOMPtr<nsIURI> sourceURI;
+
+  rv = GetDownloadDetails(getter_AddRefs(sourceURI),
+                          wideFileName);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  nsDataObj::CStream *pStream = new nsDataObj::CStream();
+  NS_ENSURE_TRUE(pStream, E_OUTOFMEMORY);
+
+  rv = pStream->Init(sourceURI);
+  if (NS_FAILED(rv))
+  {
+    pStream->Release();
+    return E_FAIL;
+  }
+  *outStream = pStream;
+
+  return S_OK;
+}
+
 ULONG nsDataObj::g_cRef = 0;
 
 EXTERN_C GUID CDECL CLSID_nsDataObj =
@@ -113,6 +373,9 @@ nsDataObj::nsDataObj(nsIURI * uri)
     // so it can create a SourceURL for CF_HTML flavour
     uri->GetSpec(mSourceURL);
   }
+
+  mIsAsyncMode = FALSE;
+  mIsInOperation = FALSE;
 }
 //-----------------------------------------------------
 // destruction
@@ -122,7 +385,7 @@ nsDataObj::~nsDataObj()
   NS_IF_RELEASE(mTransferable);
   PRInt32 i;
   for (i=0;i<mDataFlavors->Count();i++) {
-    nsCAutoString* df = NS_REINTERPRET_CAST(nsCAutoString *, mDataFlavors->ElementAt(i));
+    nsCAutoString* df = reinterpret_cast<nsCAutoString *>(mDataFlavors->ElementAt(i));
     delete df;
   }
 
@@ -131,6 +394,12 @@ nsDataObj::~nsDataObj()
   m_cRef = 0;
   m_enumFE->Release();
 
+  // Free arbitrary system formats
+  for (PRUint32 idx = 0; idx < mDataEntryList.Length(); idx++) {
+      CoTaskMemFree(mDataEntryList[idx]->fe.ptd);
+      ReleaseStgMedium(&mDataEntryList[idx]->stgm);
+      CoTaskMemFree(mDataEntryList[idx]);
+  }
 }
 
 
@@ -144,8 +413,12 @@ STDMETHODIMP nsDataObj::QueryInterface(REFIID riid, void** ppv)
 	if ( (IID_IUnknown == riid) || (IID_IDataObject	== riid) ) {
 		*ppv = this;
 		AddRef();
-		return NOERROR;
-	}
+		return S_OK;
+  } else if (IID_IAsyncOperation == riid) {
+    *ppv = static_cast<IAsyncOperation*>(this);
+    AddRef();
+    return S_OK;
+  }
 
 	return ResultFromScode(E_NOINTERFACE);
 }
@@ -168,8 +441,9 @@ STDMETHODIMP_(ULONG) nsDataObj::Release()
 	if (0 < g_cRef)
 		--g_cRef;
 
+	--m_cRef;
 	NS_LOG_RELEASE(this, m_cRef, "nsDataObj");
-	if (0 != --m_cRef)
+	if (0 != m_cRef)
 		return m_cRef;
 
 	delete this;
@@ -210,11 +484,19 @@ STDMETHODIMP nsDataObj::GetData(LPFORMATETC pFE, LPSTGMEDIUM pSTM)
   static CLIPFORMAT PreferredDropEffect = ::RegisterClipboardFormat( CFSTR_PREFERREDDROPEFFECT );
 #endif
 
+  // Arbitrary system formats
+  LPDATAENTRY pde;
+  HRESULT hres = FindFORMATETC(pFE, &pde, FALSE);
+  if (SUCCEEDED(hres)) {
+      return AddRefStgMedium(&pde->stgm, pSTM, FALSE);
+  }
+
+  // Firefox internal formats
   ULONG count;
   FORMATETC fe;
   m_enumFE->Reset();
   while (NOERROR == m_enumFE->Next(1, &fe, &count)) {
-    nsCAutoString * df = NS_REINTERPRET_CAST(nsCAutoString*, mDataFlavors->SafeElementAt(dfInx));
+    nsCAutoString * df = reinterpret_cast<nsCAutoString*>(mDataFlavors->SafeElementAt(dfInx));
     if ( df ) {
       if (FormatsMatch(fe, *pFE)) {
         pSTM->pUnkForRelease = NULL;        // caller is responsible for deleting this data
@@ -225,11 +507,7 @@ STDMETHODIMP nsDataObj::GetData(LPFORMATETC pFE, LPSTGMEDIUM pSTM)
         case CF_TEXT:
         case CF_UNICODETEXT:
         return GetText(*df, *pFE, *pSTM);
-                  
-#ifndef WINCE
-        case CF_HDROP:
-          return GetFile(*df, *pFE, *pSTM);
-#endif
+
         // Someone is asking for an image
         case CF_DIB:
           return GetDib(*df, *pFE, *pSTM);
@@ -266,7 +544,6 @@ STDMETHODIMP nsDataObj::GetData(LPFORMATETC pFE, LPSTGMEDIUM pSTM)
   return ResultFromScode(DATA_E_FORMATETC);
 }
 
-
 //-----------------------------------------------------
 STDMETHODIMP nsDataObj::GetDataHere(LPFORMATETC pFE, LPSTGMEDIUM pSTM)
 {
@@ -284,6 +561,12 @@ STDMETHODIMP nsDataObj::QueryGetData(LPFORMATETC pFE)
   PRNTDEBUG("nsDataObj::QueryGetData  ");
   PRNTDEBUG3("format: %d  Text: %d\n", pFE->cfFormat, CF_TEXT);
 
+  // Arbitrary system formats
+  LPDATAENTRY pde;
+  if (SUCCEEDED(FindFORMATETC(pFE, &pde, FALSE)))
+    return S_OK;
+
+  // Firefox internal formats
   ULONG count;
   FORMATETC fe;
   m_enumFE->Reset();
@@ -305,6 +588,33 @@ STDMETHODIMP nsDataObj::GetCanonicalFormatEtc
 		return ResultFromScode(E_FAIL);
 }
 
+HGLOBAL nsDataObj::GlobalClone(HGLOBAL hglobIn)
+{
+  HGLOBAL hglobOut = NULL;
+
+  LPVOID pvIn = GlobalLock(hglobIn);
+  if (pvIn) {
+    SIZE_T cb = GlobalSize(hglobIn);
+    HGLOBAL hglobOut = GlobalAlloc(GMEM_FIXED, cb);
+    if (hglobOut) {
+      CopyMemory(hglobOut, pvIn, cb);
+    }
+    GlobalUnlock(hglobIn);
+  }
+  return hglobOut;
+}
+
+IUnknown* nsDataObj::GetCanonicalIUnknown(IUnknown *punk)
+{
+  IUnknown *punkCanonical;
+  if (punk && SUCCEEDED(punk->QueryInterface(IID_IUnknown,
+                                             (LPVOID*)&punkCanonical))) {
+    punkCanonical->Release();
+  } else {
+    punkCanonical = punk;
+  }
+  return punkCanonical;
+}
 
 //-----------------------------------------------------
 STDMETHODIMP nsDataObj::SetData(LPFORMATETC pFE, LPSTGMEDIUM pSTM, BOOL fRelease)
@@ -322,13 +632,117 @@ STDMETHODIMP nsDataObj::SetData(LPFORMATETC pFE, LPSTGMEDIUM pSTM, BOOL fRelease
   }
 #endif
 
-  if (fRelease) {
-    ReleaseStgMedium(pSTM);
+  // Store arbitrary system formats
+  LPDATAENTRY pde;
+  HRESULT hres = FindFORMATETC(pFE, &pde, TRUE); // add
+  if (SUCCEEDED(hres)) {
+    if (pde->stgm.tymed) {
+      ReleaseStgMedium(&pde->stgm);
+      ZeroMemory(&pde->stgm, sizeof(STGMEDIUM));
+    }
+
+    if (fRelease) {
+      pde->stgm = *pSTM;
+      hres = S_OK;
+    } else {
+      hres = AddRefStgMedium(pSTM, &pde->stgm, TRUE);
+    }
+    pde->fe.tymed = pde->stgm.tymed;
+
+    // Break circular reference loop (see msdn)
+    if (GetCanonicalIUnknown(pde->stgm.pUnkForRelease) ==
+        GetCanonicalIUnknown(static_cast<IDataObject*>(this))) {
+      pde->stgm.pUnkForRelease->Release();
+      pde->stgm.pUnkForRelease = NULL;
+    }
+    return hres;
   }
 
+  if (fRelease)
+    ReleaseStgMedium(pSTM);
+
   return ResultFromScode(S_OK);
 }
 
+HRESULT
+nsDataObj::FindFORMATETC(FORMATETC *pfe, LPDATAENTRY *ppde, BOOL fAdd)
+{
+  *ppde = NULL;
+
+  if (pfe->ptd != NULL) return DV_E_DVTARGETDEVICE;
+
+  // See if it's in our list
+  for (PRUint32 idx = 0; idx < mDataEntryList.Length(); idx++) {
+    if (mDataEntryList[idx]->fe.cfFormat == pfe->cfFormat &&
+        mDataEntryList[idx]->fe.dwAspect == pfe->dwAspect &&
+        mDataEntryList[idx]->fe.lindex == pfe->lindex) {
+      if (fAdd || (mDataEntryList[idx]->fe.tymed & pfe->tymed)) {
+        *ppde = mDataEntryList[idx];
+        return S_OK;
+      } else {
+        return DV_E_TYMED;
+      }
+    }
+  }
+
+  if (!fAdd)
+    return DV_E_FORMATETC;
+
+  LPDATAENTRY pde = (LPDATAENTRY)CoTaskMemAlloc(sizeof(DATAENTRY));
+  if (pde) {
+    pde->fe = *pfe;
+    *ppde = pde;
+    ZeroMemory(&pde->stgm, sizeof(STGMEDIUM));
+
+    m_enumFE->AddFE(pfe);
+    mDataEntryList.AppendElement(pde);
+
+    return S_OK;
+  } else {
+    return E_OUTOFMEMORY;
+  }
+}
+
+HRESULT
+nsDataObj::AddRefStgMedium(STGMEDIUM *pstgmIn, STGMEDIUM *pstgmOut, BOOL fCopyIn)
+{
+  HRESULT hres = S_OK;
+  STGMEDIUM stgmOut = *pstgmIn;
+
+  if (pstgmIn->pUnkForRelease == NULL &&
+      !(pstgmIn->tymed & (TYMED_ISTREAM | TYMED_ISTORAGE))) {
+    if (fCopyIn) {
+      // Object needs to be cloned
+      if (pstgmIn->tymed == TYMED_HGLOBAL) {
+        stgmOut.hGlobal = GlobalClone(pstgmIn->hGlobal);
+        if (!stgmOut.hGlobal) {
+          hres = E_OUTOFMEMORY;
+        }
+      } else {
+        hres = DV_E_TYMED;
+      }
+    } else {
+      stgmOut.pUnkForRelease = static_cast<IDataObject*>(this);
+    }
+  }
+
+  if (SUCCEEDED(hres)) {
+    switch (stgmOut.tymed) {
+    case TYMED_ISTREAM:
+      stgmOut.pstm->AddRef();
+      break;
+    case TYMED_ISTORAGE:
+      stgmOut.pstg->AddRef();
+      break;
+    }
+    if (stgmOut.pUnkForRelease) {
+      stgmOut.pUnkForRelease->AddRef();
+    }
+    *pstgmOut = stgmOut;
+  }
+
+  return hres;
+}
 
 //-----------------------------------------------------
 STDMETHODIMP nsDataObj::EnumFormatEtc(DWORD dwDir, LPENUMFORMATETC *ppEnum)
@@ -347,16 +761,11 @@ STDMETHODIMP nsDataObj::EnumFormatEtc(DWORD dwDir, LPENUMFORMATETC *ppEnum)
         break;
   } // switch
 
-  // Since a new one has been created, 
-  // we will ref count the new clone here 
-  // before giving it back
   if (NULL == *ppEnum)
     return ResultFromScode(E_FAIL);
-  else
-    (*ppEnum)->AddRef();
 
+  // Clone already AddRefed the result so don't addref it again.
   return NOERROR;
-
 }
 
 //-----------------------------------------------------
@@ -382,6 +791,48 @@ STDMETHODIMP nsDataObj::EnumDAdvise(LPENUMSTATDATA *ppEnum)
 	return ResultFromScode(E_FAIL);
 }
 
+// IAsyncOperation methods
+STDMETHODIMP nsDataObj::EndOperation(HRESULT hResult,
+                                     IBindCtx *pbcReserved,
+                                     DWORD dwEffects)
+{
+  mIsInOperation = FALSE;
+  Release();
+  return S_OK;
+}
+
+STDMETHODIMP nsDataObj::GetAsyncMode(BOOL *pfIsOpAsync)
+{
+  if (!pfIsOpAsync)
+    return E_FAIL;
+
+  *pfIsOpAsync = mIsAsyncMode;
+
+  return S_OK;
+}
+
+STDMETHODIMP nsDataObj::InOperation(BOOL *pfInAsyncOp)
+{
+  if (!pfInAsyncOp)
+    return E_FAIL;
+
+  *pfInAsyncOp = mIsInOperation;
+
+  return S_OK;
+}
+
+STDMETHODIMP nsDataObj::SetAsyncMode(BOOL fDoOpAsync)
+{
+  mIsAsyncMode = fDoOpAsync;
+  return S_OK;
+}
+
+STDMETHODIMP nsDataObj::StartOperation(IBindCtx *pbcReserved)
+{
+  mIsInOperation = TRUE;
+  return S_OK;
+}
+
 //-----------------------------------------------------
 // other methods
 //-----------------------------------------------------
@@ -461,7 +912,7 @@ nsDataObj :: GetDib ( const nsACString& inFlavor, FORMATETC &, STGMEDIUM & aSTG
     }
   } // if we have an image
   else  
-    NS_WARNING ( "Definately not an image on clipboard" );
+    NS_WARNING ( "Definitely not an image on clipboard" );
 
 #endif
 	return ResultFromScode(result);
@@ -480,12 +931,21 @@ nsDataObj :: GetFileDescriptor ( FORMATETC& aFE, STGMEDIUM& aSTG, PRBool aIsUnic
   
   // How we handle this depends on if we're dealing with an internet
   // shortcut, since those are done under the covers.
-  if ( IsInternetShortcut() ) {
+  if (IsFlavourPresent(kFilePromiseMime) ||
+      IsFlavourPresent(kFileMime))
+  {
+    if (aIsUnicode)
+      return GetFileDescriptor_IStreamW(aFE, aSTG);
+    else
+      return GetFileDescriptor_IStreamA(aFE, aSTG);
+  }
+  else if (IsFlavourPresent(kURLMime))
+  {
     if ( aIsUnicode )
       res = GetFileDescriptorInternetShortcutW ( aFE, aSTG );
     else
       res = GetFileDescriptorInternetShortcutA ( aFE, aSTG );
-  } 
+  }
   else
     NS_WARNING ( "Not yet implemented\n" );
   
@@ -502,8 +962,11 @@ nsDataObj :: GetFileContents ( FORMATETC& aFE, STGMEDIUM& aSTG )
   
   // How we handle this depends on if we're dealing with an internet
   // shortcut, since those are done under the covers.
-  if ( IsInternetShortcut() )  
-    res = GetFileContentsInternetShortcut ( aFE, aSTG );
+  if (IsFlavourPresent(kFilePromiseMime) ||
+      IsFlavourPresent(kFileMime))
+    return GetFileContents_IStream(aFE, aSTG);
+  else if (IsFlavourPresent(kURLMime))
+    return GetFileContentsInternetShortcut ( aFE, aSTG );
   else
     NS_WARNING ( "Not yet implemented\n" );
 
@@ -604,14 +1067,13 @@ CreateFilenameFromTextW(nsString & aText, const wchar_t * aExtension,
   return PR_TRUE;
 }
 
-static NS_DEFINE_CID(kStringBundleServiceCID, NS_STRINGBUNDLESERVICE_CID);
 #define PAGEINFO_PROPERTIES "chrome://navigator/locale/pageInfo.properties"
 
 static PRBool
 GetLocalizedString(const PRUnichar * aName, nsXPIDLString & aString)
 {
   nsresult rv;
-  nsCOMPtr<nsIStringBundleService> stringService = do_GetService(kStringBundleServiceCID, &rv);
+  nsCOMPtr<nsIStringBundleService> stringService = do_GetService(NS_STRINGBUNDLE_CONTRACTID, &rv);
   if (NS_FAILED(rv)) 
     return PR_FALSE;
 
@@ -645,8 +1107,7 @@ nsDataObj :: GetFileDescriptorInternetShortcutA ( FORMATETC& aFE, STGMEDIUM& aST
   if (!fileGroupDescHandle)
     return E_OUTOFMEMORY;
 
-  LPFILEGROUPDESCRIPTORA fileGroupDescA = NS_REINTERPRET_CAST(LPFILEGROUPDESCRIPTORA, 
-      ::GlobalLock(fileGroupDescHandle));
+  LPFILEGROUPDESCRIPTORA fileGroupDescA = reinterpret_cast<LPFILEGROUPDESCRIPTORA>(::GlobalLock(fileGroupDescHandle));
   if (!fileGroupDescA) {
     ::GlobalFree(fileGroupDescHandle);
     return E_OUTOFMEMORY;
@@ -655,11 +1116,11 @@ nsDataObj :: GetFileDescriptorInternetShortcutA ( FORMATETC& aFE, STGMEDIUM& aST
   // get a valid filename in the following order: 1) from the page title, 
   // 2) localized string for an untitled page, 3) just use "Untitled.URL"
   if (!CreateFilenameFromTextA(title, ".URL", 
-          fileGroupDescA->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
+                               fileGroupDescA->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
     nsXPIDLString untitled;
     if (!GetLocalizedString(NS_LITERAL_STRING("noPageTitle").get(), untitled) ||
         !CreateFilenameFromTextA(untitled, ".URL", 
-            fileGroupDescA->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
+                                 fileGroupDescA->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
       strcpy(fileGroupDescA->fgd[0].cFileName, "Untitled.URL");
     }
   }
@@ -691,8 +1152,7 @@ nsDataObj :: GetFileDescriptorInternetShortcutW ( FORMATETC& aFE, STGMEDIUM& aST
   if (!fileGroupDescHandle)
     return E_OUTOFMEMORY;
 
-  LPFILEGROUPDESCRIPTORW fileGroupDescW = NS_REINTERPRET_CAST(LPFILEGROUPDESCRIPTORW, 
-      ::GlobalLock(fileGroupDescHandle));
+  LPFILEGROUPDESCRIPTORW fileGroupDescW = reinterpret_cast<LPFILEGROUPDESCRIPTORW>(::GlobalLock(fileGroupDescHandle));
   if (!fileGroupDescW) {
     ::GlobalFree(fileGroupDescHandle);
     return E_OUTOFMEMORY;
@@ -701,11 +1161,11 @@ nsDataObj :: GetFileDescriptorInternetShortcutW ( FORMATETC& aFE, STGMEDIUM& aST
   // get a valid filename in the following order: 1) from the page title, 
   // 2) localized string for an untitled page, 3) just use "Untitled.URL"
   if (!CreateFilenameFromTextW(title, NS_LITERAL_STRING(".URL").get(), 
-          fileGroupDescW->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
+                               fileGroupDescW->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
     nsXPIDLString untitled;
     if (!GetLocalizedString(NS_LITERAL_STRING("noPageTitle").get(), untitled) ||
         !CreateFilenameFromTextW(untitled, NS_LITERAL_STRING(".URL").get(), 
-            fileGroupDescW->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
+                                 fileGroupDescW->fgd[0].cFileName, NS_MAX_FILEDESCRIPTOR)) {
       wcscpy(fileGroupDescW->fgd[0].cFileName, NS_LITERAL_STRING("Untitled.URL").get());
     }
   }
@@ -752,7 +1212,7 @@ nsDataObj :: GetFileContentsInternetShortcut ( FORMATETC& aFE, STGMEDIUM& aSTG )
   if ( !hGlobalMemory )
     return E_OUTOFMEMORY;
 
-  char* contents = NS_REINTERPRET_CAST(char*, ::GlobalLock(hGlobalMemory));
+  char* contents = reinterpret_cast<char*>(::GlobalLock(hGlobalMemory));
   if ( !contents ) {
     ::GlobalFree( hGlobalMemory );
     return E_OUTOFMEMORY;
@@ -772,39 +1232,28 @@ nsDataObj :: GetFileContentsInternetShortcut ( FORMATETC& aFE, STGMEDIUM& aSTG )
 #endif  
 } // GetFileContentsInternetShortcut
 
-
-//
-// IsInternetShortcut
-//
-// Is the data that we're handling destined for an internet shortcut if
-// dragged to the desktop? We can tell because there will be a mime type
-// of |kURLMime| present in the transferable
-//
-PRBool
-nsDataObj :: IsInternetShortcut ( )
+// check if specified flavour is present in the transferable
+PRBool nsDataObj :: IsFlavourPresent(const char *inFlavour)
 {
   PRBool retval = PR_FALSE;
-  
-  if ( !mTransferable )
-    return PR_FALSE;
+  NS_ENSURE_TRUE(mTransferable, PR_FALSE);
   
   // get the list of flavors available in the transferable
   nsCOMPtr<nsISupportsArray> flavorList;
-  mTransferable->FlavorsTransferableCanExport ( getter_AddRefs(flavorList) );
-  if ( !flavorList )
-    return PR_FALSE;
-  
-  // go spelunking for the url flavor
+  mTransferable->FlavorsTransferableCanExport(getter_AddRefs(flavorList));
+  NS_ENSURE_TRUE(flavorList, PR_FALSE);
+
+  // try to find requested flavour
   PRUint32 cnt;
   flavorList->Count(&cnt);
-  for ( PRUint32 i = 0;i < cnt; ++i ) {
+  for (PRUint32 i = 0; i < cnt; ++i) {
     nsCOMPtr<nsISupports> genericFlavor;
     flavorList->GetElementAt (i, getter_AddRefs(genericFlavor));
     nsCOMPtr<nsISupportsCString> currentFlavor (do_QueryInterface(genericFlavor));
     if (currentFlavor) {
       nsCAutoString flavorStr;
       currentFlavor->GetData(flavorStr);
-      if ( flavorStr.Equals(kURLMime) ) {
+      if (flavorStr.Equals(inFlavour)) {
         retval = PR_TRUE;         // found it!
         break;
       }
@@ -812,8 +1261,7 @@ nsDataObj :: IsInternetShortcut ( )
   } // for each flavor
 
   return retval;
-  
-} // IsInternetShortcut
+}
 
 HRESULT nsDataObj::GetPreferredDropEffect ( FORMATETC& aFE, STGMEDIUM& aSTG )
 {
@@ -844,7 +1292,7 @@ HRESULT nsDataObj::GetPreferredDropEffect ( FORMATETC& aFE, STGMEDIUM& aSTG )
 //-----------------------------------------------------
 HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGMEDIUM& aSTG)
 {
-  void* data;
+  void* data = NULL;
   PRUint32   len;
   
   // if someone asks for text/plain, look up text/unicode instead in the transferable.
@@ -861,6 +1309,9 @@ HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGME
   if ( !len )
     return ResultFromScode(E_FAIL);
   nsPrimitiveHelpers::CreateDataFromPrimitive ( flavorStr, genericDataWrapper, &data, len );
+  if ( !data )
+    return ResultFromScode(E_FAIL);
+
   HGLOBAL     hGlobalMemory = NULL;
 
   aSTG.tymed          = TYMED_HGLOBAL;
@@ -879,7 +1330,7 @@ HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGME
     // Someone is asking for text/plain; convert the unicode (assuming it's present)
     // to text with the correct platform encoding.
     char* plainTextData = nsnull;
-    PRUnichar* castedUnicode = NS_REINTERPRET_CAST(PRUnichar*, data);
+    PRUnichar* castedUnicode = reinterpret_cast<PRUnichar*>(data);
     PRInt32 plainTextLen = 0;
     nsPrimitiveHelpers::ConvertUnicodeToPlatformPlainText ( castedUnicode, len / 2, &plainTextData, &plainTextLen );
    
@@ -898,7 +1349,7 @@ HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGME
   else if ( aFE.cfFormat == nsClipboard::CF_HTML ) {
     // Someone is asking for win32's HTML flavor. Convert our html fragment
     // from unicode to UTF-8 then put it into a format specified by msft.
-    NS_ConvertUCS2toUTF8 converter ( NS_REINTERPRET_CAST(PRUnichar*, data) );
+    NS_ConvertUTF16toUTF8 converter ( reinterpret_cast<PRUnichar*>(data) );
     char* utf8HTML = nsnull;
     nsresult rv = BuildPlatformHTML ( converter.get(), &utf8HTML );      // null terminates
     
@@ -923,8 +1374,8 @@ HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGME
 
   // Copy text to Global Memory Area
   if ( hGlobalMemory ) {
-    char* dest = NS_REINTERPRET_CAST(char*, GlobalLock(hGlobalMemory));
-    char* source = NS_REINTERPRET_CAST(char*, data);
+    char* dest = reinterpret_cast<char*>(GlobalLock(hGlobalMemory));
+    char* source = reinterpret_cast<char*>(data);
     memcpy ( dest, source, allocLen );                         // copies the null as well
     GlobalUnlock(hGlobalMemory);
   }
@@ -936,76 +1387,6 @@ HRESULT nsDataObj::GetText(const nsACString & aDataFlavor, FORMATETC& aFE, STGME
   return ResultFromScode(S_OK);
 }
 
-//-----------------------------------------------------
-HRESULT nsDataObj::GetFile(const nsACString & aDataFlavor, FORMATETC& aFE, STGMEDIUM& aSTG)
-{
-  HRESULT res = S_OK;
-  if (strcmp(PromiseFlatCString(aDataFlavor).get(), kFilePromiseMime) == 0) {
-    // XXX We are copying the file twice below.  Once from the original location to the temporary
-    // one (here) and then from the temporary location to the drop location (done by the drop target
-    // when this function returns).  If the file being drag-dropped is in the cache, we should be able to
-    // just pass back its file name to the drop target and save a copy.  Need help here!
-    nsresult rv;
-    if (!mCachedTempFile) {
-      // Save the file to a temporary location.      
-      nsCOMPtr<nsILocalFile> dropDirectory;
-      nsCOMPtr<nsIProperties> directoryService = 
-        do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID, &rv);
-      if (!directoryService || NS_FAILED(rv)) return ResultFromScode(E_FAIL);
-      directoryService->Get(NS_OS_TEMP_DIR, 
-                            NS_GET_IID(nsIFile), 
-                            getter_AddRefs(dropDirectory));
-      nsCOMPtr<nsISupports> localFileISupports = do_QueryInterface(dropDirectory);
-      rv = mTransferable->SetTransferData(kFilePromiseDirectoryMime, localFileISupports, sizeof(nsILocalFile*));
-      if (NS_FAILED(rv)) return ResultFromScode(E_FAIL);
-      nsCOMPtr<nsISupports> fileDataPrimitive;
-      PRUint32 dataSize = 0;
-      rv = mTransferable->GetTransferData(kFilePromiseMime, getter_AddRefs(fileDataPrimitive), &dataSize);
-      if (NS_FAILED(rv)) return ResultFromScode(E_FAIL);      
-      
-      mCachedTempFile = do_QueryInterface(fileDataPrimitive);
-      if (!mCachedTempFile) return ResultFromScode(E_FAIL);
-    }
-
-    // Pass the file name back to the drop target so that it can copy to the drop location
-    nsAutoString path;
-    rv = mCachedTempFile->GetPath(path);
-    if (NS_FAILED(rv)) return ResultFromScode(E_FAIL);
-    // Two null characters are needed to terminate the file name list
-    PRUint32 allocLen = path.Length() + 2;
-    HGLOBAL hGlobalMemory = NULL;
-    aSTG.tymed = TYMED_HGLOBAL;
-    aSTG.pUnkForRelease = NULL;
-    hGlobalMemory = GlobalAlloc(GMEM_MOVEABLE, sizeof(DROPFILES) + allocLen * sizeof(PRUnichar));
-    if (hGlobalMemory) {
-      DROPFILES* pDropFile = NS_REINTERPRET_CAST(DROPFILES*, GlobalLock(hGlobalMemory));
-
-      // First, populate drop file structure
-      pDropFile->pFiles = sizeof(DROPFILES);   // offset to start of file name char array
-      pDropFile->fNC = 0;
-      pDropFile->pt.x = 0;
-      pDropFile->pt.y = 0;
-      pDropFile->fWide = TRUE;
-
-      // Copy the filename right after the DROPFILES structure
-      PRUnichar* dest = NS_REINTERPRET_CAST(PRUnichar*, NS_REINTERPRET_CAST(char*, pDropFile) + pDropFile->pFiles);
-      memcpy(dest, path.get(), (allocLen - 1) * sizeof(PRUnichar));    // copies the null character in path as well
-
-      // Two null characters are needed at the end of the file name.  
-      // Lookup the CF_HDROP shell clipboard format for more info.
-      // Add the second null character right after the first one.
-      dest[allocLen - 1] = L'\0';
-
-      GlobalUnlock(hGlobalMemory);
-    }
-    else {
-      res = E_OUTOFMEMORY;
-    }
-    aSTG.hGlobal = hGlobalMemory;
-  }
-  return res;
-}
-
 //-----------------------------------------------------
 HRESULT nsDataObj::GetMetafilePict(FORMATETC&, STGMEDIUM&)
 {
@@ -1019,7 +1400,7 @@ HRESULT nsDataObj::SetBitmap(FORMATETC&, STGMEDIUM&)
 }
 
 //-----------------------------------------------------
-HRESULT nsDataObj::SetDib   (FORMATETC&, STGMEDIUM&)
+HRESULT nsDataObj::SetDib(FORMATETC&, STGMEDIUM&)
 {
 	return ResultFromScode(E_FAIL);
 }
@@ -1045,7 +1426,7 @@ HRESULT nsDataObj::SetText  (FORMATETC& aFE, STGMEDIUM& aSTG)
 }
 
 //-----------------------------------------------------
-HRESULT nsDataObj::SetMetafilePict (FORMATETC&, STGMEDIUM&)
+HRESULT nsDataObj::SetMetafilePict(FORMATETC&, STGMEDIUM&)
 {
 	return ResultFromScode(E_FAIL);
 }
@@ -1069,14 +1450,11 @@ void nsDataObj::AddDataFlavor(const char* aDataFlavor, LPFORMATETC aFE)
   // so we will look up data flavor that corresponds to the FE
   // and then ask the transferable for that type of data
 
-  // Add CF_HDROP to the head of the list so that it is returned first
-  // when the drop target calls EnumFormatEtc to enumerate through the FE list
-  // We use the CF_HDROP format to copy file contents when an image is dragged
-  // from Mozilla to a drop target.
+  // Just ignore the CF_HDROP here
+  // all file drags are now hangled by CFSTR_FileContents format
 #ifndef WINCE
   if (aFE->cfFormat == CF_HDROP) {
-    mDataFlavors->InsertElementAt(new nsCAutoString(aDataFlavor), 0);
-    m_enumFE->InsertFEAt(aFE, 0);
+    return;
   }  
   else 
 #endif
@@ -1300,7 +1678,7 @@ HRESULT
 nsDataObj :: GetUniformResourceLocator( FORMATETC& aFE, STGMEDIUM& aSTG, PRBool aIsUnicode )
 {
   HRESULT res = S_OK;
-  if ( IsInternetShortcut() ) {
+  if (IsFlavourPresent(kURLMime)) {
     if ( aIsUnicode )
       res = ExtractUniformResourceLocatorW( aFE, aSTG );
     else
@@ -1326,7 +1704,7 @@ nsDataObj::ExtractUniformResourceLocatorA(FORMATETC& aFE, STGMEDIUM& aSTG )
   if (!hGlobalMemory)
     return E_OUTOFMEMORY;
 
-  char* contents = NS_REINTERPRET_CAST(char*, GlobalLock(hGlobalMemory));
+  char* contents = reinterpret_cast<char*>(GlobalLock(hGlobalMemory));
   if (!contents) {
     GlobalFree(hGlobalMemory);
     return E_OUTOFMEMORY;
@@ -1354,7 +1732,7 @@ nsDataObj::ExtractUniformResourceLocatorW(FORMATETC& aFE, STGMEDIUM& aSTG )
   if (!hGlobalMemory)
     return E_OUTOFMEMORY;
 
-  wchar_t* contents = NS_REINTERPRET_CAST(wchar_t*, GlobalLock(hGlobalMemory));
+  wchar_t* contents = reinterpret_cast<wchar_t*>(GlobalLock(hGlobalMemory));
   if (!contents) {
     GlobalFree(hGlobalMemory);
     return E_OUTOFMEMORY;
@@ -1367,3 +1745,141 @@ nsDataObj::ExtractUniformResourceLocatorW(FORMATETC& aFE, STGMEDIUM& aSTG )
 
   return result;
 }
+
+
+// Gets the filename from the kFilePromiseURLMime flavour
+nsresult nsDataObj::GetDownloadDetails(nsIURI **aSourceURI,
+                                       nsAString &aFilename)
+{
+  *aSourceURI = nsnull;
+
+  NS_ENSURE_TRUE(mTransferable, NS_ERROR_FAILURE);
+
+  // get the URI from the kFilePromiseURLMime flavor
+  nsCOMPtr<nsISupports> urlPrimitive;
+  PRUint32 dataSize = 0;
+  mTransferable->GetTransferData(kFilePromiseURLMime, getter_AddRefs(urlPrimitive), &dataSize);
+  nsCOMPtr<nsISupportsString> srcUrlPrimitive = do_QueryInterface(urlPrimitive);
+  NS_ENSURE_TRUE(srcUrlPrimitive, NS_ERROR_FAILURE);
+  
+  nsAutoString srcUri;
+  srcUrlPrimitive->GetData(srcUri);
+  if (srcUri.IsEmpty())
+    return NS_ERROR_FAILURE;
+  nsCOMPtr<nsIURI> sourceURI;
+  NS_NewURI(getter_AddRefs(sourceURI), srcUri);
+
+  nsAutoString srcFileName;
+  nsCOMPtr<nsISupports> fileNamePrimitive;
+  mTransferable->GetTransferData(kFilePromiseDestFilename, getter_AddRefs(fileNamePrimitive), &dataSize);
+  nsCOMPtr<nsISupportsString> srcFileNamePrimitive = do_QueryInterface(fileNamePrimitive);
+  if (srcFileNamePrimitive) {
+    srcFileNamePrimitive->GetData(srcFileName);
+  } else {
+    nsCOMPtr<nsIURL> sourceURL = do_QueryInterface(sourceURI);
+    if (!sourceURL)
+      return NS_ERROR_FAILURE;
+    
+    nsCAutoString urlFileName;
+    sourceURL->GetFileName(urlFileName);
+    NS_UnescapeURL(urlFileName);
+    CopyUTF8toUTF16(urlFileName, srcFileName);
+  }
+  if (srcFileName.IsEmpty())
+    return NS_ERROR_FAILURE;
+
+  // make the name safe for the filesystem
+  MangleTextToValidFilename(srcFileName);
+
+  sourceURI.swap(*aSourceURI);
+  aFilename = srcFileName;
+  return NS_OK;
+}
+
+HRESULT nsDataObj::GetFileDescriptor_IStreamA(FORMATETC& aFE, STGMEDIUM& aSTG)
+{
+  HGLOBAL fileGroupDescHandle = ::GlobalAlloc(GMEM_ZEROINIT|GMEM_SHARE,sizeof(FILEGROUPDESCRIPTORW));
+  NS_ENSURE_TRUE(fileGroupDescHandle, E_OUTOFMEMORY);
+
+  LPFILEGROUPDESCRIPTORA fileGroupDescA = reinterpret_cast<LPFILEGROUPDESCRIPTORA>(GlobalLock(fileGroupDescHandle));
+  if (!fileGroupDescA) {
+    ::GlobalFree(fileGroupDescHandle);
+    return E_OUTOFMEMORY;
+  }
+
+  nsAutoString wideFileName;
+  nsresult rv;
+  nsCOMPtr<nsIURI> sourceURI;
+  rv = GetDownloadDetails(getter_AddRefs(sourceURI),
+                          wideFileName);
+  if (NS_FAILED(rv))
+  {
+    ::GlobalFree(fileGroupDescHandle);
+    return E_FAIL;
+  }
+
+  nsCAutoString nativeFileName;
+  NS_UTF16ToCString(wideFileName, NS_CSTRING_ENCODING_NATIVE_FILESYSTEM, nativeFileName);
+  
+  strncpy(fileGroupDescA->fgd[0].cFileName, nativeFileName.get(), NS_MAX_FILEDESCRIPTOR - 1);
+  fileGroupDescA->fgd[0].cFileName[NS_MAX_FILEDESCRIPTOR - 1] = '\0';
+
+  // one file in the file block
+  fileGroupDescA->cItems = 1;
+  fileGroupDescA->fgd[0].dwFlags = FD_PROGRESSUI;
+
+  GlobalUnlock( fileGroupDescHandle );
+  aSTG.hGlobal = fileGroupDescHandle;
+  aSTG.tymed = TYMED_HGLOBAL;
+
+  return S_OK;
+}
+
+HRESULT nsDataObj::GetFileDescriptor_IStreamW(FORMATETC& aFE, STGMEDIUM& aSTG)
+{
+  HGLOBAL fileGroupDescHandle = ::GlobalAlloc(GMEM_ZEROINIT|GMEM_SHARE,sizeof(FILEGROUPDESCRIPTORW));
+  NS_ENSURE_TRUE(fileGroupDescHandle, E_OUTOFMEMORY);
+
+  LPFILEGROUPDESCRIPTORW fileGroupDescW = reinterpret_cast<LPFILEGROUPDESCRIPTORW>(GlobalLock(fileGroupDescHandle));
+  if (!fileGroupDescW) {
+    ::GlobalFree(fileGroupDescHandle);
+    return E_OUTOFMEMORY;
+  }
+
+  nsAutoString wideFileName;
+  nsresult rv;
+  nsCOMPtr<nsIURI> sourceURI;
+  rv = GetDownloadDetails(getter_AddRefs(sourceURI),
+                          wideFileName);
+  if (NS_FAILED(rv))
+  {
+    ::GlobalFree(fileGroupDescHandle);
+    return E_FAIL;
+  }
+
+  wcsncpy(fileGroupDescW->fgd[0].cFileName, wideFileName.get(), NS_MAX_FILEDESCRIPTOR - 1);
+  fileGroupDescW->fgd[0].cFileName[NS_MAX_FILEDESCRIPTOR - 1] = '\0';
+  // one file in the file block
+  fileGroupDescW->cItems = 1;
+  fileGroupDescW->fgd[0].dwFlags = FD_PROGRESSUI;
+
+  GlobalUnlock(fileGroupDescHandle);
+  aSTG.hGlobal = fileGroupDescHandle;
+  aSTG.tymed = TYMED_HGLOBAL;
+
+  return S_OK;
+}
+
+HRESULT nsDataObj::GetFileContents_IStream(FORMATETC& aFE, STGMEDIUM& aSTG)
+{
+  IStream *pStream = NULL;
+
+  nsDataObj::CreateStream(&pStream);
+  NS_ENSURE_TRUE(pStream, E_FAIL);
+
+  aSTG.tymed = TYMED_ISTREAM;
+  aSTG.pstm = pStream;
+  aSTG.pUnkForRelease = pStream;
+
+  return S_OK;
+}
